/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamconnection.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"
#include "gdamevent.h"
#include "gdamcachebuffer.h"
#include "gdamserverdebug.h"
#include <unistd.h>
#include <errno.h>

GdamConnection* gdam_connection_for_arg = NULL;

static gboolean    gdam_connection_prepare        (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data);
static gboolean    gdam_connection_check          (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);
static gboolean    gdam_connection_dispatch       (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);

static GSourceFuncs gdam_connection_source_funcs = {
	gdam_connection_prepare,
	gdam_connection_check,
	gdam_connection_dispatch,
	NULL
};

GdamConnection* gdam_connection_new(int client_fd,
                                    GdamAuthenticator* authenticator)
{
	GdamConnection* connection;

	if (!authenticator) {
		g_warning("no authenticator for connection:  cannot accept");
		return NULL;
	}
	connection = g_new0(GdamConnection, 1);
	gdam_buffer_construct(&connection->client_commands);
	connection->authenticator = authenticator;
	connection->client_poll.events = G_IO_IN;
	connection->client_poll.fd = client_fd;
	connection->source_id = 0;
	gdam_buffer_construct(&connection->outgoing);
	connection->is_readable = 1;
	connection->ref_table = g_hash_table_new(NULL, NULL);
	gdam_permissions_none(&connection->permissions);

	return connection;
}

static void do_unreferences(gpointer key, gpointer value, gpointer con)
{
	GtkObject* object = key;
	int count = GPOINTER_TO_INT(value);
	while (count-- > 0)
		gtk_object_unref(object);
}

static void gdam_connection_destroy(GdamConnection* connection)
{
	gdam_event_disconnect_all_from_connection(connection);

	gdam_debug("gdam_connection_destroy: %d references to clear",
	  connection->ref_table ? g_hash_table_size(connection->ref_table) : 0);
		

	if (connection->ref_table) {
		g_hash_table_foreach(connection->ref_table,
				     do_unreferences,
				     connection);
		g_hash_table_destroy(connection->ref_table);
		connection->ref_table = NULL;
	}

	if (connection->client_poll.fd >= 0) {
		g_main_remove_poll(&connection->client_poll);
		close(connection->client_poll.fd);
		connection->client_poll.fd = -1;
	}
	if (connection->source_id) {
		guint source_id = connection->source_id;
		connection->source_id = 0;
		g_source_remove(source_id);
	}
	gdam_buffer_destruct(&connection->client_commands);
	gdam_buffer_destruct(&connection->outgoing);
	g_free(connection);
}

static void dispatch_disconnected(gpointer conn)
{
	GdamConnection* connection = (GdamConnection*)conn;
	if (connection->client_poll.fd >= 0) {
		g_main_remove_poll(&connection->client_poll);
		close(connection->client_poll.fd);
		connection->client_poll.fd = -1;
	}
	connection->source_id = 0;
	gdam_connection_destroy(connection);
}

void gdam_connection_attach_to_main_loop(GdamConnection* connection)
{
	g_return_if_fail(connection->source_id == 0);
	g_return_if_fail(connection->client_poll.fd >= 0);

	connection->source_id = g_source_add(G_PRIORITY_DEFAULT,
	                                     FALSE,	/* cannot recurse */
					     &gdam_connection_source_funcs,
					     NULL,/* no source specific data */
					     connection,
					     dispatch_disconnected);
	g_main_add_poll(&connection->client_poll, G_PRIORITY_DEFAULT);
}

static gboolean gdam_connection_prepare(gpointer src_data,
                                        GTimeVal* curtime,
                                        gint* timeout,
		                        gpointer user_data)
{
	GdamConnection* connection = (GdamConnection*)user_data;

	connection->client_poll.revents = 0;
	*timeout = -1;

	/* Set up event flag... */
	if (connection->is_readable)
		connection->client_poll.events = G_IO_IN|G_IO_HUP|G_IO_ERR;
	else
		connection->client_poll.events = 0;
	if (connection->outgoing.size)
		connection->client_poll.events |= G_IO_OUT;

	/* We always need an event to do anything useful... */
	return FALSE;
}

static gboolean gdam_connection_check(gpointer         src_data,
                                      GTimeVal*        curtime,
		                      gpointer         user_data)
{
	GdamConnection* connection = (GdamConnection*)user_data;
	if (connection->client_poll.fd < 0) {
		if (GDAM_SERVER_DEBUG_POLL)
			gdam_debug("gdam_connection_check: no fd");
		return FALSE;
	}
	if (connection->client_poll.revents == 0) {
		if (GDAM_SERVER_DEBUG_POLL)
			gdam_debug("gdam_connection_check: no revents");
		return FALSE;
	}
	if (GDAM_SERVER_DEBUG_POLL)
		gdam_debug("gdam_connection_check: ready");
	return TRUE;
}

static inline GdamBuffer* do_dispatch(GdamConnection* connection)
{
	GdamBuffer command;
	GdamBuffer* rv;
	int length;
	gdam_buffer_peek_int(&connection->client_commands, &length);
	gdam_buffer_construct(&command);
	/*gdam_debug("do_dispatch:  starting size=%d, length=%d",
	                        connection->client_commands.size, length);*/
	gdam_buffer_send_buffer(&command,
	                        &connection->client_commands,
				length,
				TRUE);
	g_assert(command.size == length);
	/*gdam_debug("do_dispatch:  remaining client input's length=%d",
	                        connection->client_commands.size);*/
		
	gdam_buffer_parse_int(&command, &length);

	if (connection->authenticator) {
		GdamAuthenticator* auth = connection->authenticator;
		gboolean is_done;
		gdam_debug("doing authentification");
		rv = (*auth->authenticate)(&command,
		                           &connection->permissions,
					   &is_done);
		if (is_done) {
			if (auth->destroy)
				(*auth->destroy)(auth);
			connection->authenticator = NULL;
		}
	} else {
		gdam_namespace_local_clean();
		rv = gdam_dispatch_buffer(connection,
		                          &connection->permissions,
					  &command);
	}
	gdam_buffer_destruct(&command);
	gdam_buffer_prepend_length(rv);
	return rv;
}

static gboolean gdam_connection_dispatch(gpointer         src_data,
                                         GTimeVal*        curtime,
		                         gpointer         user_data)
{
	GdamConnection* connection = (GdamConnection*)user_data;

	/* The events returned from the poll() system-call. */
	int revents = connection->client_poll.revents;

	int client_fd = connection->client_poll.fd;

	g_return_val_if_fail(client_fd >= 0, FALSE);

	if (revents & G_IO_IN) {
		char buf[16*1024];
		/* The buffer for client input. */
		GdamBuffer* client_buf = &connection->client_commands;
		int num_read;

		/* Do the reading. (and error handling) */
		num_read = read(client_fd, buf, sizeof(buf));
		if (num_read < 0) {
			if (errno == EINTR
			 || errno == EWOULDBLOCK
			 || errno == EINPROGRESS) {
			 	return TRUE;
			}
			g_warning("dispatch command:  reading from client: %s",
				strerror(errno));
			connection->source_id = 0;
			return FALSE;
		}

		/* Add newly-read data into "incoming" buffer. */
		gdam_buffer_append(client_buf, buf, num_read);
		/*gdam_debug("appending %d bytes from client [new size=%d]",
				num_read, client_buf->size);*/

		if (num_read == 0) {
			/* end-of-file condition... */
			connection->is_readable = 0;
		} else {
			GdamBuffer* response;
			int cmd_len=0;
			while (gdam_buffer_peek_int(client_buf, &cmd_len)
			 && cmd_len <= client_buf->size) {
#if DEBUG_DISPATCHING
				gdam_debug("dispatching message... len=%d",
					cmd_len);
#endif
				response = do_dispatch(connection);
				if (!response) {
					g_warning("error processing command");
					connection->source_id = 0;
					return FALSE;
				}
				gdam_buffer_append_buffer(
					&connection->outgoing,
					response,
					TRUE
				);
				gdam_buffer_destroy(response);
			}
#if DEBUG_DISPATCHING
			if (gdam_buffer_peek_int(client_buf, &cmd_len))
				gdam_debug("waiting to dispatch message: "
				           "cur_len=%d, cmd_len=%d",
					   client_buf->size, cmd_len);
#endif
		}
	}

	/* Handle outgoing data. */
	if (revents & G_IO_OUT && connection->outgoing.size > 0) {
		if (!gdam_buffer_write_one(&connection->outgoing, client_fd)) {
			connection->source_id = 0;
			return FALSE;
		}
	}

	if (connection->outgoing.size == 0
	 && !connection->is_readable) {
		connection->source_id = 0;
	 	return FALSE;
	}

	if (revents & (G_IO_HUP|G_IO_ERR))
		return FALSE;

	return TRUE;
}

void  gdam_connection_emit_event(GdamConnection* connection,
                                 GdamBuffer*     event)
{
	gdam_buffer_prepend_length(event);
	gdam_buffer_append_buffer(&connection->outgoing, event, TRUE);
}

gboolean gdam_connection_ref(GdamConnection* connection, GtkObject* object)
{
	int old_count;
	g_return_val_if_fail(object != NULL, FALSE);
	old_count = (int) (g_hash_table_lookup(connection->ref_table, object));
	gtk_object_ref(object);
	if (old_count == 0)
		gtk_object_sink(object);
	g_hash_table_insert(connection->ref_table,
	                    object, 
	                    GINT_TO_POINTER(old_count + 1));
	return TRUE;
}

gboolean gdam_connection_unref(GdamConnection* connection, GtkObject* object)
{
	int old_count;
	old_count = (int) (g_hash_table_lookup(connection->ref_table, object));
	if (old_count <= 0) {
		gdam_debug("tried to unref too many times");
		return FALSE;
	}
	gtk_object_unref(object);
	g_hash_table_insert(connection->ref_table,
	                    object, 
	                    GINT_TO_POINTER(old_count - 1));
	return TRUE;
}
