/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsynthdelaylineallpass.h"
#include <gdam/gdamplugin.h>

static int gdam_synth_delay_line_all_pass_synthesize
				(GdamSynthSource*      source,
				 gdouble*              output,
				 int                   num_read)
{
	GdamSynthSource* this_filtered = GDAM_SYNTH_FILTER(source)->filtered;
	GdamSynthDelayLineAllPass* this_all_pass;
	GdamSynthDelayLine* this_delay_line;
	gdouble* inputs;
	int in_point, out_point, buffer_length;
	gdouble last_in, last_out, coeff;
	int n;
	this_all_pass = GDAM_SYNTH_DELAY_LINE_ALL_PASS(source);
	this_delay_line = GDAM_SYNTH_DELAY_LINE(source);
	if (!this_filtered)
		return 0;
	n = gdam_synth_source_synthesize(this_filtered, output, num_read);
	if (n <= 0)
		return 0;
	num_read = n;
	inputs = this_delay_line->buffer;
	in_point = this_delay_line->in_point;
	out_point = this_delay_line->out_point;
	coeff = this_all_pass->coeff;
	last_out = this_all_pass->last_out;
	last_in = this_all_pass->last_in;
	buffer_length = this_delay_line->buffer_length;
	while (num_read-- > 0) {
		gdouble tmp;
		inputs[in_point++] = *output;
		if (in_point == buffer_length)
			in_point = 0;
		tmp = inputs[out_point++];
		if (out_point == buffer_length)
			out_point = 0;
		last_out = -coeff * last_out;
		last_out += last_in + coeff * tmp;
		last_in = tmp;
		*output++ = last_out;
	}
	this_delay_line->in_point = in_point;
	this_delay_line->out_point = out_point;
	this_all_pass->last_out = last_out;
	this_all_pass->last_in = last_in;
	return n;
}

static void gdam_synth_delay_line_all_pass_init
                                      (GdamSynthDelayLineAllPass* synth)
{
	int buffer_length = 2048;
	synth->out_point_fraction = 0.0;
	gdam_synth_delay_line_set_buffer_length(GDAM_SYNTH_DELAY_LINE(synth),
	                                        buffer_length);
}

static void
gdam_synth_delay_line_all_pass_class_init(GdamSynthSourceClass* klass)
{
	klass->synthesize = gdam_synth_delay_line_all_pass_synthesize;
}

void gdam_synth_delay_line_all_pass_set_delay
			(GdamSynthDelayLineAllPass* source,
			 gdouble                    lag)
{
	gdouble out_pointer;
	gdouble alpha, coeff;
	GdamSynthDelayLine* delay_line = GDAM_SYNTH_DELAY_LINE(source);
	GdamSynthDelayLineAllPass* all_pass;
	int out_point, in_point;
	int length = delay_line->buffer_length;

	all_pass = GDAM_SYNTH_DELAY_LINE_ALL_PASS(source);
	if (lag > length - 1) {
		/* TODO: fix this. */
		gdam_debug("lag too long");
		out_pointer = in_point - 18.0;
	} else if (lag < 0.1) {
		gdam_debug("DLineA: Delays < 0.1 not possible ... "
                           "Setting delay length to 0.1");
		out_pointer = in_point + 0.8999999999;
	} else {
		out_pointer = in_point - lag + 1.0;
	}

	if (out_pointer < 0)
		out_pointer += length;
	out_point = (long) out_pointer;    /* Integer part of delay */
	alpha = 1.0 + out_point - out_pointer;  /* fractional part */

	if (alpha == 0.0) {  		/* Exact integer delay. */
		out_point -= 1;
		if (out_point < 0) out_point += length;
	}

	/* Hack to avoid pole/zero cancellation.  Keeps allpass
	   delay in range of .1 to 1.1 */
	if (alpha < 0.1) {
		out_point += 1;                             
		if (out_point >= length) out_point -= length;
		alpha += 1.0;
	}

	coeff = (1.0 - alpha) / (1.0 + alpha);

	delay_line->out_point = out_point;
	all_pass->out_point_fraction = alpha;
	delay_line->in_point = in_point;
	source->coeff = coeff;
}

GtkType gdam_synth_delay_line_all_pass_get_type()
{
	static GtkType synth_delay_line_all_pass_type = 0;
	if (!synth_delay_line_all_pass_type) {
		GtkType parent;
		static const GtkTypeInfo synth_delay_line_all_pass_info = {
			"GdamSynthDelayLineAllPass",
			sizeof(GdamSynthDelayLineAllPass),
			sizeof(GdamSynthDelayLineAllPassClass),
			(GtkClassInitFunc)
				gdam_synth_delay_line_all_pass_class_init,
			(GtkObjectInitFunc)
				gdam_synth_delay_line_all_pass_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent = gdam_plugin_get_type_by_name("source",
		                                      "GdamSynthDelayLine");
		synth_delay_line_all_pass_type = gtk_type_unique(parent,
					 &synth_delay_line_all_pass_info);
	}
	return synth_delay_line_all_pass_type;
}
