/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamvolumemeter.h"

static void gdam_volume_meter_init(GdamVolumeMeter* volume_meter)
{
	volume_meter->window = 8192;
	gdam_buffering_filter_ensure_space(GDAM_BUFFERING_FILTER(volume_meter),
	                                   volume_meter->window);
}

static void gdam_volume_meter_arg_get_window(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_VOLUME_METER(object)->window;
}
static void gdam_volume_meter_arg_set_window(GtkObject* object, GdamArg* arg)
{
	GdamVolumeMeter* meter = GDAM_VOLUME_METER(object);
	meter->window = GDAM_ARG_INT(arg);
	gdam_buffering_filter_ensure_space(GDAM_BUFFERING_FILTER(meter),
	                                   meter->window);
}

#define SAMPLE_TO_INT(sample)     ((int) ( (sample).s[0] ))

static void gdam_volume_meter_get_ranges(GdamVolumeMeter* volume_meter,
                                         int*             left_min_out,
                                         int*             left_max_out,
                                         int*             right_min_out,
                                         int*             right_max_out)
{
	GdamBufferingFilter* buffering;
	int buflen;
	int left_min, right_min, left_max, right_max;

	buffering = GDAM_BUFFERING_FILTER(volume_meter);

	buflen = buffering->buffer_current_amount;
	if (buflen == 0) {
		left_max = left_min = 0;
		right_max = right_min = 0;
	} else {
		GdamSample sample;
		int i;

		sample = gdam_buffering_filter_sample_at(buffering, 0);
		left_max = left_min = sample.s[0];
		right_max = right_min = sample.s[1];

		for (i = 1; i < buflen; i++) {
			int left, right;
			sample = gdam_buffering_filter_sample_at(buffering, i);
			left = sample.s[0];
			right = sample.s[1];
			if (left < left_min)
				left_min = left;
			else if (left > left_max)
				left_max = left;
			if (right < right_min)
				right_min = right;
			else if (right > right_max)
				right_max = right;
		}
	}
	if (left_min_out != NULL)
		*left_min_out = left_min;
	if (left_max_out != NULL)
		*left_max_out = left_max;
	if (right_min_out != NULL)
		*right_min_out = right_min;
	if (right_max_out != NULL)
		*right_max_out = right_max;
}
static void gdam_volume_meter_arg_get_range(GtkObject* object, GdamArg* arg)
{
	int* min_max;
	min_max = g_new(int, 4);
	GDAM_ARG_INT_ARRAY(arg).length = 4;
	GDAM_ARG_INT_ARRAY(arg).array = min_max;
	gdam_volume_meter_get_ranges(GDAM_VOLUME_METER(object),
	                             min_max + 0,
	                             min_max + 1,
	                             min_max + 2,
	                             min_max + 3);
}


static void gdam_volume_meter_class_init(GtkObjectClass* object_class)
{
	gdam_arg_register(object_class,
	                  "window",
			  GDAM_ARG_TYPE_INT,
			  gdam_volume_meter_arg_set_window,
			  gdam_volume_meter_arg_get_window,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "range",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  NULL,
			  gdam_volume_meter_arg_get_range,
			  GDAM_ARG_VITAL);
}


GtkType gdam_volume_meter_get_type()
{
	static GtkType volume_meter_type = 0;
	if (!volume_meter_type) {
		static const GtkTypeInfo volume_meter_info = {
			"GdamVolumeMeter",
			sizeof(GdamVolumeMeter),
			sizeof(GdamVolumeMeterClass),
			(GtkClassInitFunc) gdam_volume_meter_class_init,
			(GtkObjectInitFunc) gdam_volume_meter_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_BUFFERING_FILTER;
		volume_meter_type = gtk_type_unique(parent,
		                                 &volume_meter_info);
	}
	return volume_meter_type;
}
