/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include <gtk/gtksignal.h>
#include <stdlib.h>
#include "gdammidiselector.h"
#include "../../component/plugins/gdammididevice.h"
#include "../gdamskinutil.h"

enum {
	DONE_GROPING,
	LAST_SIGNAL
};

static guint midi_selector_signals[LAST_SIGNAL] = { 0 };

static void gdam_midi_selector_class_init (GdamSkinClass* skin_class);

static void gdam_midi_selector_init(GdamMidiSelector* midi_selector)
{
}
GtkType gdam_midi_selector_get_type()
{
	static GtkType midi_selector_type = 0;
	if (!midi_selector_type) {
		static const GtkTypeInfo midi_selector_info = {
			"GdamMidiSelector",
			sizeof(GdamMidiSelector),
			sizeof(GdamMidiSelectorClass),
			(GtkClassInitFunc) gdam_midi_selector_class_init,
			(GtkObjectInitFunc) gdam_midi_selector_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		midi_selector_type = gtk_type_unique(parent,
		                                 &midi_selector_info);
	}
	return midi_selector_type;
}


static GdamMidiSelector* latest = NULL;
static guint             latest_destroy_id = 0;

static void set_latest_midi_selector(GdamMidiSelector* selector)
{
	if (latest == selector)
		return;
	if (latest != NULL) {
		gtk_signal_disconnect(GTK_OBJECT(latest),
		                      latest_destroy_id);
		latest = NULL;
	}
	if (selector != NULL) {
		/* This is certainly too tricky and clever... */
		latest_destroy_id =
			gtk_signal_connect_object(GTK_OBJECT(selector),
					          "destroy",
				  GTK_SIGNAL_FUNC(set_latest_midi_selector),
					          NULL);
		latest = selector;
	}
}

static GdamMidiSelector* get_latest_midi_selector()
{
	return latest;
}
static gboolean gdam_midi_selector_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GDAM_MIDI_SELECTOR(skin)->entry = GTK_ENTRY(object);
	return TRUE;
}

static void midi_selector_open(GdamMidiSelector* skin)
{
	char* device_name;
	GdamMidiDevice* midi_device;
	g_return_if_fail(skin->entry != NULL);
	device_name = gtk_entry_get_text(skin->entry);
	midi_device = gdam_midi_device_get_unique(device_name);
}

static gboolean gdam_midi_selector_open_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	set_latest_midi_selector(GDAM_MIDI_SELECTOR(skin));
	gtk_signal_connect_object(object,
	                          signal_name,
				  GTK_SIGNAL_FUNC(midi_selector_open),
				  (GtkObject*) skin);
	return TRUE;
}

/* FIXME:  the columns are hardcoded.
 *             0 is the patch name,
 *             1 is the description,
 *             2 is the destination.
 */

static void handle_new_trap(GdamMidiSelector* selector, GdamMidiTrap* trap)
{
	/* XXX: destination hasn't been dealt with. */
	const char* description = trap->trap_name;
	//const char* device_name = trap->device->filename;
	char* trap_id;
	char* strings[3];

	int index;

	if (selector->clist == NULL)
		return;

	trap_id = gdam_midi_trap_get_string(trap);
	strings[0] = trap_id;
	strings[1] = (char*) description;
	strings[2] = "";
	index = gtk_clist_append(selector->clist, strings);
	gtk_clist_set_row_data(selector->clist, index, trap);
	g_free(trap_id);
}
static gboolean clist_emission_hook(GtkObject*       object,
                                    guint            signal_id,
				    guint            n_params,
				    GtkArg*          params,
				    gpointer         skin_as_data)
{
	GdamMidiSelector* skin = GDAM_MIDI_SELECTOR(skin_as_data);
	GdamMidiTrap* trap;
	/* GdamMidiEvent* event; */
	g_return_val_if_fail(skin != NULL, TRUE);
	g_return_val_if_fail(n_params == 2, TRUE);
	g_return_val_if_fail(params[0].type == GTK_TYPE_POINTER
	                 &&  params[1].type == GTK_TYPE_POINTER, TRUE);
	trap = (GdamMidiTrap*) GTK_VALUE_POINTER(params[0]);
	/* event = (GdamMidiEvent*) GTK_ARG_POINTER(params[1]); */
	handle_new_trap(skin, trap);
	return TRUE;
}

static void handle_midi_iterate(GdamMidiTrap* trap, gpointer user_data)
{
	handle_new_trap(GDAM_MIDI_SELECTOR(user_data), trap);
}

/* Attach this handler in the xml file to register this
 * clist with the skin. */

static void patch_selected(GtkWidget*      clist,
                           gint            row,
			   gint            col,
			   GdkEventButton* event,
                           GdamMidiSelector* selector)
{
	GdamMidiTrap* trap;
	trap = gtk_clist_get_row_data(selector->clist, row);

	selector->selected_device = trap->device;
	selector->selected_event_type = trap->type;
	selector->selected_channel = trap->channel;
	selector->selected_which_param = trap->which_param;
}

static gboolean gdam_midi_selector_clist_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	guint new_trap_id;
	GdamMidiSelector* sel_skin = GDAM_MIDI_SELECTOR(skin);
	GtkCList* clist = GTK_CLIST(object);
	g_return_val_if_fail(sel_skin->clist == NULL, TRUE);
	g_return_val_if_fail(clist != NULL, TRUE);
	new_trap_id = gtk_signal_lookup("new-trap",
	                                GDAM_TYPE_MIDI_DEVICE);
	g_return_val_if_fail(new_trap_id != 0, TRUE);
	sel_skin->clist = clist;
	sel_skin->emission_id = gtk_signal_add_emission_hook
	                                            (new_trap_id,
	                                             clist_emission_hook,
				                     skin);
	gtk_signal_connect(GTK_OBJECT(clist),
	                   "select_row",
			   GTK_SIGNAL_FUNC(patch_selected),
			   sel_skin);
	                   
	gtk_clist_freeze(clist);
	gdam_midi_device_iterate_traps(NULL, handle_midi_iterate, skin);
	gtk_clist_thaw(clist);
	return TRUE;
}

/* XXX: we just assume model will not be destroyed,
        since we are calling it from a skin. */
typedef struct _DoMidiPatch DoMidiPatch;
struct _DoMidiPatch {
	int ref_count;
	GdamModel* model;
	char* arg_name;
	char* description;
	gdouble min_val, max_val;
	int done_groping_connect_id;
};

static void do_midi_patch_unref(gpointer patch)
{
	DoMidiPatch* midi_patch = patch;
	g_return_if_fail(midi_patch->ref_count > 0);
	if (--(midi_patch->ref_count) == 0) {
		g_return_if_fail(midi_patch->done_groping_connect_id == 0);
		g_free(patch);
	}
}

static void handle_midi_hook( GdamMidiDevice*    midi_dev,
	                      GdamMidiEvent*     midi_event,
			      gpointer           user_data)
{
	DoMidiPatch* patch = user_data;
	GdamArg arg;
	int type;
	gdouble norm, val;
	arg.name = patch->arg_name;
	switch (midi_event->type) {
		case GDAM_MIDI_PITCH_WHEEL: {
			norm = midi_event->pitch_event.norm_value;
			break;
		}
		case GDAM_MIDI_PARAMETER: {
			norm = midi_event->param_event.norm_value;
			break;
		}
		default: {
			g_warning("sorry, the midi event of internal type %d"
			          " cannot be trapped yet", midi_event->type);
			return;
		}
	}
	val = norm * (patch->max_val - patch->min_val) + patch->min_val;
	type = gdam_arg_query_type(GTK_OBJECT(patch->model)->klass->type, 
					      patch->arg_name);

	if (type == GDAM_ARG_TYPE_DOUBLE){
		arg.type = GDAM_ARG_TYPE_DOUBLE;
		GDAM_ARG_DOUBLE(&arg) = val;
	}
	else if (type == GDAM_ARG_TYPE_INT){
		arg.type = GDAM_ARG_TYPE_INT;
		GDAM_ARG_INT(&arg) = (int)val;
	}
	else {
		g_warning("type mismatch in midiselector handle hook");
	}
	gdam_arg_set_object_arg(GTK_OBJECT(patch->model), &arg);
}

void gdam_midi_selector_stop_groping(GdamMidiSelector* selector);
void gdam_midi_selector_start_groping(GdamMidiSelector* selector);

/* User has de-activated the 'grope' state. */
static void do_midi_grope_stop(){
	GdamMidiSelector* midi_sel = get_latest_midi_selector();
	g_return_if_fail(midi_sel != NULL);
	gdam_debug("doing midi_grope_stop.");
	gdam_midi_selector_stop_groping(midi_sel);
}

static void on_done_groping(GdamMidiSelector* selector,
			    DoMidiPatch*      patch_data)
{
	GdamMidiSelector* midi_sel = get_latest_midi_selector();
	GdamMidiTrap* trap;
	g_return_if_fail(midi_sel != NULL);
	if (midi_sel->grope == NULL){
		gdam_midi_device_disconnect_by_type(NULL,
				            GTK_OBJECT(patch_data->model),
					    patch_data->description);
		/* TODO: this assumes latest_midi_sel is unchanged */
		gtk_signal_disconnect(GTK_OBJECT(midi_sel), 
				      patch_data->done_groping_connect_id);
		patch_data->done_groping_connect_id = 0;
		return;
	}
	trap = midi_sel->grope;
	patch_data->ref_count++;
		/* TODO: this assumes trap->device unchanged */
	gdam_midi_device_disconnect_by_type(trap->device,
				            GTK_OBJECT(patch_data->model),
					    patch_data->description);
	gdam_midi_device_connect(trap->device,
	                         trap->type,
				 trap->channel,
				 trap->which_param,
				 patch_data->description,
				 handle_midi_hook,
				 do_midi_patch_unref,
				 patch_data,
				 GTK_OBJECT(patch_data->model));
	/* TODO: this assumes latest_midi_sel is unchanged from connection */
	gdam_debug("done groping,  disconnecting done_groping_id.");
	gtk_signal_disconnect(GTK_OBJECT(midi_sel), 
			      patch_data->done_groping_connect_id);
	patch_data->done_groping_connect_id = 0;
}

static void do_midi_patch_clear(DoMidiPatch* patch_data){
	gdam_midi_device_disconnect_by_type(NULL,
				            GTK_OBJECT(patch_data->model),
					    patch_data->description);
}

/* User has activated the 'grope' state. */
static void do_midi_grope_start(DoMidiPatch* patch_data)
{
	GdamMidiSelector* midi_sel = get_latest_midi_selector();
	int id;
	g_return_if_fail(midi_sel != NULL);
	g_return_if_fail(patch_data->done_groping_connect_id == 0);
	gdam_debug("doing midi_grope_start.");
	gdam_midi_selector_start_groping(midi_sel);
	patch_data->ref_count++;
	id = gtk_signal_connect_full(GTK_OBJECT(midi_sel),
	                             "done_groping",
				     GTK_SIGNAL_FUNC(on_done_groping),
				     NULL,
				     patch_data,
				     do_midi_patch_unref,
				     FALSE,
				     FALSE);
	patch_data->done_groping_connect_id = id;
}

static void do_midi_patch(DoMidiPatch* patch_data)
{
	GdamMidiSelector* midi_sel = get_latest_midi_selector();
	g_return_if_fail(midi_sel != NULL);
	if (midi_sel->selected_device == NULL) {
		g_warning("before patching midi, you must select a patch "
		          "from the Midi Selector");
		return;
	}
	patch_data->ref_count++;
	gdam_midi_device_connect(midi_sel->selected_device,
	                         midi_sel->selected_event_type,
				 midi_sel->selected_channel,
				 midi_sel->selected_which_param,
				 patch_data->description,
				 handle_midi_hook,
				 do_midi_patch_unref,
				 patch_data,
				 GTK_OBJECT(patch_data->model));
}

/* A midi grope for a generic skin.
 * (The skin here is *not* a midi selector)
 */
static gboolean gdam_midi_grope_stop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_debug("gdam_midi_grope_stop handled.");
	gtk_signal_connect(object, 
			   signal_name,
			   do_midi_grope_stop,
			   NULL);
	return TRUE;
}


/* A midi grope for a generic skin.
 * (The skin here is *not* a midi selector)
 */
static gboolean gdam_midi_patch_clear_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	char* arg_name;
	char* desc;
	char* endp;
	gdouble min_val, max_val;
	DoMidiPatch* midi_patch;
	if (num_handler_names != 5) {
		g_warning("usage: gdam_midi_patch_clear:argname:min:max:description");
		return TRUE;
	}
	arg_name = g_strdup(handler_names[1]);
	min_val = strtod(handler_names[2], &endp);
	if (*endp != '\0') {
		g_warning("midi_patch: error parsing double");
		return TRUE;
	}
	max_val = strtod(handler_names[3], &endp);
	if (*endp != '\0') {
		g_warning("midi_patch: error parsing double");
		return TRUE;
	}
	desc = g_strdup(handler_names[4]);

	midi_patch = g_new(DoMidiPatch, 1);
	midi_patch->ref_count = 1;
	midi_patch->model = skin->model;
	midi_patch->arg_name  = arg_name;
	midi_patch->description = desc;
	midi_patch->min_val = min_val;
	midi_patch->max_val = max_val;
	midi_patch->done_groping_connect_id = 0;
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(do_midi_patch_clear),
				NULL,
				midi_patch,
				do_midi_patch_unref,
				TRUE,
				FALSE);
	gdam_debug("gdam_midi_patch_clear handled.");
	return TRUE;
}

/* A midi grope for a generic skin.
 * (The skin here is *not* a midi selector)
 */
static gboolean gdam_midi_grope_start_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	char* arg_name;
	char* desc;
	char* endp;
	gdouble min_val, max_val;
	DoMidiPatch* midi_patch;
	if (num_handler_names != 5) {
		g_warning("usage: gdam_midi_grope:argname:min:max:description");
		return TRUE;
	}
	arg_name = g_strdup(handler_names[1]);
	min_val = strtod(handler_names[2], &endp);
	if (*endp != '\0') {
		g_warning("midi_patch: error parsing double");
		return TRUE;
	}
	max_val = strtod(handler_names[3], &endp);
	if (*endp != '\0') {
		g_warning("midi_patch: error parsing double");
		return TRUE;
	}
	desc = g_strdup(handler_names[4]);

	midi_patch = g_new(DoMidiPatch, 1);
	midi_patch->ref_count = 1;
	midi_patch->model = skin->model;
	midi_patch->arg_name  = arg_name;
	midi_patch->description = desc;
	midi_patch->min_val = min_val;
	midi_patch->max_val = max_val;
	midi_patch->done_groping_connect_id = 0;
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(do_midi_grope_start),
				NULL,
				midi_patch,
				do_midi_patch_unref,
				TRUE,
				FALSE);
	gdam_debug("gdam_midi_grope_start handled.");
	return TRUE;
}

/* A midi patch for a generic skin.
 * when signalled, attaches to current device and event
 * (The skin here is *not* a midi selector)
 *
 * Arguments:
 *       argname            : the name of the double-argument to change
 *       min_val            : the minimum value for that param
 *       max_val            : the maximum value for that param
 *       description        : midi hook description
 */
static gboolean gdam_midi_patch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	char* arg_name;
	char* desc;
	char* endp;
	gdouble min_val, max_val;
	DoMidiPatch* midi_patch;
	if (num_handler_names != 5) {
		g_warning("usage: gdam_midi_patch:argname:min:max:description");
		return TRUE;
	}
	arg_name = g_strdup(handler_names[1]);
	min_val = strtod(handler_names[2], &endp);
	if (*endp != '\0') {
		g_warning("midi_patch: error parsing double");
		return TRUE;
	}
	max_val = strtod(handler_names[3], &endp);
	if (*endp != '\0') {
		g_warning("midi_patch: error parsing double");
		return TRUE;
	}
	desc = g_strdup(handler_names[4]);

	midi_patch = g_new(DoMidiPatch, 1);
	midi_patch->ref_count = 1;
	midi_patch->model = skin->model;
	midi_patch->arg_name  = arg_name;
	midi_patch->description = desc;
	midi_patch->min_val = min_val;
	midi_patch->max_val = max_val;
	midi_patch->done_groping_connect_id = 0;
	gtk_signal_connect_full(object,
	                        signal_name,
				GTK_SIGNAL_FUNC(do_midi_patch),
				NULL,
				midi_patch,
				do_midi_patch_unref,
				TRUE,
				FALSE);
	return TRUE;
}

static GdamSkinGenericConnect connect_infos[] =
{ 	{ "gdam_midi_selector_open",
	   gdam_midi_selector_open_handler },
	{ "gdam_midi_selector_entry",
	   gdam_midi_selector_entry_handler },
	{ "gdam_midi_selector_clist",
	   gdam_midi_selector_clist_handler }, };

static GdamSkinGenericConnect skin_connect_info[] =
{	{ "gdam_midi_patch",
           gdam_midi_patch_handler },
 	{ "gdam_midi_grope_stop",
           gdam_midi_grope_stop_handler },
 	{ "gdam_midi_grope_start",
           gdam_midi_grope_start_handler },
 	{ "gdam_midi_patch_clear",
           gdam_midi_patch_clear_handler } };

static void register_skin_connect_infos()
{
	GdamSkinClass* skin_class;
	skin_class = GDAM_SKIN_CLASS(gtk_type_class(GDAM_TYPE_SKIN));
	gdam_skin_class_register_generic_handlers
					(skin_class,
					 G_N_ELEMENTS(skin_connect_info),
					 skin_connect_info);
}

static void gdam_midi_selector_class_init (GdamSkinClass* skin_class)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(skin_class);
	midi_selector_signals[DONE_GROPING]
		= gtk_signal_new("done_groping",
		                 GTK_RUN_LAST,
                                 object_class->type,
				 GTK_SIGNAL_OFFSET(GdamMidiSelectorClass,
				                   done_groping),
				 gtk_marshal_NONE__NONE,
				 GTK_TYPE_NONE,
				 0);
	gtk_object_class_add_signals(object_class,
	                             midi_selector_signals,
				     LAST_SIGNAL);
	gdam_skin_class_register_generic_handlers(skin_class,
				                  G_N_ELEMENTS(connect_infos),
				                  connect_infos);
	register_skin_connect_infos();
}

static gboolean grope_midi_event_handler(GtkObject*       object,
                                         guint            signal_id,
				         guint            n_params,
				         GtkArg*          params,
				         gpointer         skin_as_data)
{
	GdamMidiTrap* trap;
	trap = gdam_midi_device_get_trap
	                        (GDAM_MIDI_DEVICE(object),
				 (GdamMidiEvent*)GTK_VALUE_POINTER(params[0]),
				 FALSE);
	g_return_val_if_fail(trap != NULL, TRUE);
	gdam_debug("grope_midi_event found trap.");
	GDAM_MIDI_SELECTOR(skin_as_data)->grope = trap;
	return TRUE;
}

void gdam_midi_selector_start_groping(GdamMidiSelector* selector)
{
	guint event_id;
	g_return_if_fail(selector->grope_emission_id == 0);
	event_id = gtk_signal_lookup("event", GDAM_TYPE_MIDI_DEVICE);
	g_return_if_fail(event_id != 0);
	gdam_debug("midi_selector_start_groping.");
	selector->grope_emission_id = gtk_signal_add_emission_hook
	                                            (event_id,
	                                             grope_midi_event_handler,
				                     selector);
}

void gdam_midi_selector_stop_groping(GdamMidiSelector* selector)
{
	guint event_id;
	g_return_if_fail(selector->grope_emission_id != 0);
	event_id = gtk_signal_lookup("event", GDAM_TYPE_MIDI_DEVICE);
	g_return_if_fail(event_id != 0);
	gtk_signal_remove_emission_hook(event_id, selector->grope_emission_id);
	selector->grope_emission_id = 0;
	gtk_signal_emit(GTK_OBJECT(selector),
			midi_selector_signals[DONE_GROPING]);
	if (selector->grope == NULL)
		gdam_debug("midi_selector_stop_groping, NO TRAP");
	selector->grope = NULL;
}
